%  Copyright (C) 2002-2004 David Roundy
%  Copyright (C) 2004 Juliusz Chroboczek
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software Foundation,
%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

\begin{code}
module Current ( Current,
                 createCurrent, removeCurrent, identifyCurrent,
                 checkCurrent, slurpCurrent,
                 writeCurrent, write_dirty_Current,
                 syncCurrent, replaceCurrent, getCurrentPop
               ) where

import Maybe ( isJust, fromMaybe )
import Monad ( when, liftM )
import Directory ( createDirectory, doesDirectoryExist, doesFileExist,
                   renameDirectory )
import Lock ( rm_recursive, writeBinFile )
import Diff ( cmp, sync )
import Workaround ( getCurrentDirectory )
import SlurpDirectory ( Slurpy,  mmap_slurp, slurp_write, slurp_write_dirty)
import DarcsUtils ( withCurrentDirectory, catchall )

import PatchInfo ( PatchInfo )
import PopulationData ( Population, getPopFrom )


data Current = NoCurrent | PlainCurrent | HashedCurrent

identifyCurrent :: IO (Current)
identifyCurrent = do mc <- reallyIdentifyCurrent
                     case mc of
                         Nothing -> fail "_darcs/current doesn't exist"
                         Just cur -> return cur

reallyIdentifyCurrent :: IO (Maybe Current)
reallyIdentifyCurrent = do direxists <- doesDirectoryExist "_darcs/current"
                           noexists <- doesFileExist "_darcs/current.none"
                           hashexists <- doesFileExist "_darcs/current.hash"
                           case (direxists, noexists, hashexists) of
                             (False, False, False) -> return Nothing
                             (True, False, False) -> return (Just PlainCurrent)
                             (False, True, False) -> return (Just NoCurrent)
                             (False, False, True) -> 
                                 return (Just HashedCurrent)
                             _ -> fail "Multiple current representations."

createCurrent :: (Maybe Current) -> IO Current
createCurrent mc = do oldcur <- reallyIdentifyCurrent
                      when (isJust oldcur) $
                          fail "_darcs/current already exists."
                      case (fromMaybe PlainCurrent mc) of
                          NoCurrent -> do writeBinFile "_darcs/current.none"
                                              "Do not delete this file.\n"
                                          return NoCurrent
                          PlainCurrent -> do createDirectory "_darcs/current"
                                             return PlainCurrent
                          HashedCurrent -> 
                              fail "HashedCurrent is not implemented yet."

removeCurrent :: Current -> IO ()
removeCurrent NoCurrent = return ()
removeCurrent PlainCurrent = rm_recursive "_darcs/current"
removeCurrent HashedCurrent = fail "HashedCurrent is not implemented yet."

checkCurrent :: FilePath -> Current -> IO Bool
checkCurrent _ NoCurrent = return True
checkCurrent path PlainCurrent = do cwd <- getCurrentDirectory
                                    cmp (cwd ++ "/_darcs/current") path
checkCurrent _ HashedCurrent = fail "HashedCurrent is not implemented yet."

slurpCurrent :: Current -> IO (Maybe Slurpy)
slurpCurrent PlainCurrent = do cwd <- getCurrentDirectory
                               slurpy <- mmap_slurp $
                                   cwd ++ "/_darcs/current"
                               return (Just slurpy)
slurpCurrent _ = return Nothing

writeCurrent :: Slurpy -> Current -> IO ()
writeCurrent _ NoCurrent = return ()
writeCurrent cur PlainCurrent =
    withCurrentDirectory "_darcs/current" $
        slurp_write cur
writeCurrent _ HashedCurrent = 
    fail "HashedCurrent is not implemented yet."

write_dirty_Current :: Slurpy -> Current -> IO ()
write_dirty_Current _ NoCurrent = return ()
write_dirty_Current cur PlainCurrent = 
    withCurrentDirectory "_darcs/current" $
        slurp_write_dirty cur
write_dirty_Current _ HashedCurrent = 
    fail "HashedCurrent is not implemented yet."
     

syncCurrent :: Slurpy -> Slurpy -> Current -> IO ()
syncCurrent _ _ NoCurrent = return ()
syncCurrent ocur owork PlainCurrent = sync "_darcs/current" ocur owork
syncCurrent _ _ HashedCurrent = return ()

replaceCurrent :: FilePath -> Current -> IO ()
replaceCurrent _ NoCurrent = return ()
replaceCurrent newcur PlainCurrent =
    do rm_recursive "_darcs/current_old"
           `catchall` return ()
       renameDirectory "_darcs/current" "_darcs/current_old"
       renameDirectory newcur "_darcs/current"
       return ()
replaceCurrent _ HashedCurrent = fail "HashedCurrent is not implemented yet."

getCurrentPop :: PatchInfo -> Current -> IO (Maybe Population)
getCurrentPop pinfo PlainCurrent = 
    Just `liftM` getPopFrom "_darcs/current" pinfo
getCurrentPop _ _ = return Nothing
\end{code}
